import numpy
import cv2


class Stitcher(object):
    def __init__(self, pictures):
        self.img_list = pictures
        self.w = 80

    @staticmethod
    def add_padding(img, w):
        h, _, _ = img.shape
        padding = numpy.zeros((h, w, 3), numpy.uint8) + 210
        return numpy.hstack((img, padding))

    @staticmethod
    def merge_with_param(img1, img2, w, roi_scale, tail_scale, index):
        img1 = img1.copy()
        img2 = img2.copy()
        h1, w1, _ = img1.shape
        h2, w2, _ = img2.shape
        img1 = img1[0:h1 - int(tail_scale * h2), :, ]
        if w1 == w2:
            img1 = Stitcher.add_padding(img1, w)
        h1, w1, _ = img1.shape
        _img1 = img1[:, :w1 - w, :]
        h1, w1, _ = _img1.shape
        if w1 != w2:
            raise Exception("Image merge: different width")
        roi_y = int(h2 * roi_scale)
        roi = _img1[(h1 - roi_y):h1, :, :]
        res = cv2.matchTemplate(img2, roi, cv2.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        std = numpy.std(roi)
        if std < 10:
            max_val = 0.95
        cut_point = (max_loc[0], max_loc[1] + int(roi_scale * h2))
        img2 = Stitcher.add_padding(img2, w)
        img = numpy.vstack((img1, img2[cut_point[1]:h2, :, :]))
        cv2.putText(img, "-" + str(index), (w2, h1), cv2.FONT_ITALIC, 1.5, (123, 189, 60), 5)
        return img, max_val

    @staticmethod
    def stack_image(img1, img2, w, index):
        img1 = img1.copy()
        img2 = img2.copy()
        h1, w1, _ = img1.shape
        h2, w2, _ = img2.shape
        if w1 == w2:
            img1 = Stitcher.add_padding(img1, w)
        h1, w1, _ = img1.shape
        _img1 = img1[:, :w1 - w, :]
        h1, w1, _ = _img1.shape
        if w1 != w2:
            raise Exception("Image width different")
        img2 = Stitcher.add_padding(img2, w)
        img = numpy.vstack((img1, img2))
        cv2.putText(img, "-" + str(index), (w2, h1), cv2.FONT_ITALIC, 1.5, (123, 189, 60), 5)
        return img

    @staticmethod
    def img_merge(img1, img2, index, w, merge=True):
        match = 0.98
        min_match = 0.92
        img_list = []
        score_list = []
        scale_list = [
            {
                "roi_scale": 0.12,
                "tail_scale": 0.18
            }, {
                "roi_scale": 0.08,
                "tail_scale": 0.32
            }, {
                "roi_scale": 0.08,
                "tail_scale": 0.08
            }, {
                "roi_scale": 0.05,
                "tail_scale": 0.2
            }, {
                "roi_scale": 0.1,
                "tail_scale": 0.4
            }, {
                "roi_scale": 0.08,
                "tail_scale": 0.15
            }
        ]
        if merge:
            for scale in scale_list:
                img, score = Stitcher.merge_with_param(img1, img2, w, scale["roi_scale"], scale["tail_scale"], index)
                img_list.append(img)
                score_list.append(score)
                if score > match:
                    break
                if scale == scale_list[-1] and max(score_list) < min_match:
                    img = Stitcher.stack_image(img1, img2, w, index)
                else:
                    img = img_list[score_list.index(max(score_list))]
        else:
            img = Stitcher.stack_image(img1, img2, w, index)
        return img

    def image_merge(self, name, without_padding, merge=True):
        """
        :param name: image path to save after merge
        :param merge: operate image merge
        :param without_padding: remove padding on the merge image
        :return: image merge name
        """
        img_list = []
        w = 0 if without_padding else self.w
        for img in self.img_list:
            img_list.append('capture/'+img)
        name = 'capture/'+name
        if len(img_list) < 2:
            cv2.imwrite(name, cv2.imread(img_list[0]))
        else:
            img1 = img_list[0]
            for img in img_list[1:]:
                index = img_list.index(img)
                if img_list.index(img) == 1:
                    img1 = self.img_merge(cv2.imread(img1), cv2.imread(img), index, w, merge)
                else:
                    img1 = self.img_merge(img1, cv2.imread(img), index, w, merge)
            img_merge = img1
            cv2.imwrite(name, img_merge)
        return name



import cv2
import numpy


class HashSimilar(object):
    @staticmethod
    def perception_hash(img_gray, precision=64):
        """
        get hash code of image
        """
        img_scale = cv2.resize(img_gray, (precision, precision))
        img_list = img_scale.flatten()
        avg = sum(img_list)*1./len(img_list)
        avg_list = ['0' if i < avg else '1' for i in img_list]
        return [int(''.join(avg_list[x:x+4]), 2) for x in range(0, precision*precision)]

    @staticmethod
    def get_image(img_file):
        """
        get image in numpy shape
        """
        img = cv2.imread(img_file)
        h, w, _ = img.shape
        img = img[int(w * 0.078):, :, :]
        img_gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        return img_gray

    @staticmethod
    def get_image_list(img_gray):
        """
        get image in list shape
        """
        i = 0
        img_list = []
        h, w = img_gray.shape
        stride = int(w*0.05)
        while i < h:
            img_list.append(img_gray[i:i + stride, :])
            i = i + stride
        return img_list

    @staticmethod
    def hamming_dist(s1, s2):
        assert len(s1) == len(s2)
        return sum([ch1 != ch2 for ch1, ch2 in zip(s1, s2)])

    @staticmethod
    def get_attention(img1, img2):
        """
        calculate images attention score list
        :param img1: input image A
        :param img2: input image B
        :return: score list
        """
        img1_list = HashSimilar.get_image_list(img1)
        img2_list = HashSimilar.get_image_list(img2)
        l = min(len(img1_list), len(img2_list))
        score_list = []
        for i in range(0, l, 1):
            start = i - 5 if i - 5 >= 0 else 0
            img_stack = numpy.vstack(img2_list[start:i+5])
            res = cv2.matchTemplate(img1_list[i], img_stack, cv2.TM_CCOEFF_NORMED)
            min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
            score_list.append(max_val)
        return score_list

    @staticmethod
    def get_attention_similar(image1, image2) -> float:
        """
        calculate image content similar
        :param image1: input image A
        :param image2: input image B
        :return: similar score
            - 0.2 : image come from different schema
            - 0.8 : image has increment content
            - 1.0 : imageA is similar to imageB
        """
        img1 = HashSimilar.get_image(image1)
        img2 = HashSimilar.get_image(image2)
        std1 = numpy.std(img1)
        std2 = numpy.std(img2)
        img1_list = HashSimilar.get_image_list(img1)
        img2_list = HashSimilar.get_image_list(img2)
        l = min(len(img1_list), len(img2_list))
        score_list = HashSimilar.get_attention(img1, img2)
        score_list.sort()
        if score_list[int(len(score_list) * 0.8) - 1] < 0.8 or max(len(img1_list), len(img2_list)) - l > l \
                or abs(std1 - std2) > 35:
            return 0.2
        if min(score_list) > 0.99:
            return 1.0
        return 0.8

    @staticmethod
    def get_hash_similar(image1, image2) -> float:
        """
        :param image1: input image A
        :param image2: input image B
        :return: score of hash calculation
        """
        img1 = HashSimilar.get_image(image1)
        img2 = HashSimilar.get_image(image2)
        hash1 = HashSimilar.perception_hash(img1)
        hash2 = HashSimilar.perception_hash(img2)
        score = 1 - HashSimilar.hamming_dist(hash1, hash2) * 1.0 / (64 * 64)
        return score



import hashlib
import os
from flask import jsonify
from flask import request
from flask import Blueprint
from flask import make_response
from visionx.service.image_diff import ImageDiff
from visionx.service.image_infer import get_ui_infer
from visionx.service.image_merge import Stitcher
from visionx.service.image_similar import HashSimilar
from visionx.service.image_text import get_image_text
from visionx.service.image_trace import image_trace
from visionx.service.image_utils import download_image, get_pop_v, save_base64_image


vision = Blueprint('vision', __name__, url_prefix='/vision')


@vision.route('/diff', methods=["POST"])
def vision_diff():
    data = {
        "code": 0,
        "data": ImageDiff().get_image_score(request.json['image1'], request.json['image2'],
                                            request.json['image_diff_name'])
    }
    return jsonify(data)


@vision.route('/merge', methods=["POST"])
def vision_merge():
    data = {
        "code": 0,
        "data": Stitcher(request.json['image_list']).image_merge(
            request.json['name'],
            without_padding=request.json.get('without_padding')
        )
    }
    return jsonify(data)


@vision.route('/similar', methods=["POST"])
def vision_similar():
    data = {
        "code": 0,
        "data": HashSimilar().get_hash_similar(request.json['image1'], request.json['image2'])
    }
    return jsonify(data)


@vision.route('/pop', methods=["POST"])
def vision_pop():
    data = {
        "code": 0,
        "data": get_pop_v(request.json['image'])
    }
    return jsonify(data)


@vision.route('/text', methods=["POST"])
def vision_text():
    data = {
        "code": 0,
        "data": get_image_text(request.json['image'])
    }
    resp = make_response(jsonify(data))
    resp.headers["Content-Type"] = "application/json;charset=utf-8"
    return resp


@vision.route('/ui-infer', methods=["POST"])
def vision_infer():
    code = 0
    image_type = request.json.get('type', 'url')
    cls_thresh = request.json.get('cls_thresh', 0.5)
    if image_type == 'url':
        img_url = request.json['url']
        image_name = f'{hashlib.md5(img_url.encode(encoding="utf-8")).hexdigest()}.{img_url.split(".")[-1]}'
        image_path = download_image(img_url, image_name)
    elif image_type == 'base64':
        base64_image = request.json['image']
        image_name = f'{hashlib.md5(base64_image.encode(encoding="utf-8")).hexdigest()}.png'
        image_path = save_base64_image(base64_image, image_name)
    else:
        raise Exception(f'UI infer API don`t support this type: {image_type}')

    try:
        data = get_ui_infer(image_path, cls_thresh)
    finally:
        os.remove(image_path)

    result = {
        "code": code,
        "data": data
    }
    resp = make_response(jsonify(result))
    resp.headers["Content-Type"] = "application/json;charset=utf-8"
    return resp


@vision.route('/semantic-search', methods=["POST"])
def semantic_search():
    """
    req json:
    {
        type: [optional] 'url' or 'base64', 'url'
        target_image,
        source_image,
        target_desc,
        image_alpha [float] 0.0-1.0,
        text_alpha: [float] 0.0-1.0,
        top_k: [optional] integer value >=1, 1
        proposal_provider: [optional] 'ui-infer' or 'patches', 'ui-infer'
    }
    """
    code = 0
    image_type = request.json.get('type', 'url')
    target_image = request.json.get('target_image')
    source_image = request.json.get('source_image')
    if image_type == 'url':
        image_name = f'{hashlib.md5(target_image.encode(encoding="utf-8")).hexdigest()}.{target_image.split(".")[-1]}'
        target_image_path = download_image(target_image, image_name)
        image_name = f'{hashlib.md5(source_image.encode(encoding="utf-8")).hexdigest()}.{source_image.split(".")[-1]}'
        source_image_path = download_image(source_image, image_name)
    elif image_type == 'base64':
        image_name = f'{hashlib.md5(target_image.encode(encoding="utf-8")).hexdigest()}.png'
        target_image_path = save_base64_image(target_image, image_name)
        image_name = f'{hashlib.md5(source_image.encode(encoding="utf-8")).hexdigest()}.png'
        source_image_path = save_base64_image(source_image, image_name)
    else:
        raise Exception(f'Not supported type: {image_type}')
    try:
        target_image_info = {'img': target_image_path, 'desc': request.json.get('target_desc', '')}
        top_k_ids, scores, infer_result, max_confidence = image_trace.search_image(
            target_image_info, source_image_path, request.json.get('top_k', 1), request.json.get('image_alpha'),
            request.json.get('text_alpha'), request.json.get('proposal_provider', 'ui-infer')
        )
        data = []
        for i in top_k_ids:
            data.append({
                'score': round(float(scores[i]), 2),
                'boxes': [int(k) for k in infer_result[i]['elem_det_region']]
            })
    finally:
        os.remove(target_image_path)
        os.remove(source_image_path)
    result = {
        "code": code,
        "data": {
            'max_confidence': max_confidence,
            'search_result': data
        }
    }
    resp = make_response(jsonify(result))
    resp.headers["Content-Type"] = "application/json;charset=utf-8"
    return resp



from flask import Flask
from flask import jsonify
from flask_cors import CORS
from visionx.api.vision_api import vision

app = Flask(__name__)
CORS(app)

app.register_blueprint(vision, url_prefix='/vision')


@app.errorhandler(Exception)
def error(e):
    ret = dict()
    ret["code"] = 1
    ret["data"] = repr(e)
    return jsonify(ret)


def main():
    app.run(host="0.0.0.0", port=9092)


if __name__ == '__main__':
    main()



import base64
import os
import cv2
import numpy
import requests
from visionx.config import IMAGE_TEMP_DIR


def merge_rectangle_contours(rectangle_contours):
    merged_contours = [rectangle_contours[0]]
    for rec in rectangle_contours[1:]:
        for i in range(len(merged_contours)):
            x_min = rec[0][0]
            y_min = rec[0][1]
            x_max = rec[2][0]
            y_max = rec[2][1]
            merged_x_min = merged_contours[i][0][0]
            merged_y_min = merged_contours[i][0][1]
            merged_x_max = merged_contours[i][2][0]
            merged_y_max = merged_contours[i][2][1]
            if x_min >= merged_x_min and y_min >= merged_y_min and x_max <= merged_x_max and y_max <= merged_y_max:
                break
            else:
                if i == len(merged_contours)-1:
                    merged_contours.append(rec)
    return merged_contours


def contour_area_filter(binary, contours, thresh=1500):
    rectangle_contours =[]
    h, w = binary.shape
    for contour in contours:
        if thresh < cv2.contourArea(contour) < 0.2*h*w:
            rectangle_contours.append(contour)
    return rectangle_contours


def get_roi_image(img, rectangle_contour):
    roi_image = img[rectangle_contour[0][1]:rectangle_contour[2][1],
                    rectangle_contour[0][0]:rectangle_contour[1][0]]
    return roi_image


def get_infer_area(infer_result):
    area = 0
    for result in infer_result:
        region = result['elem_det_region']
        area = area + (int((region[2] - region[0]))*(int(region[3] - region[1])))
    return area


def get_pop_v(image):
    """
    calculate value if a pop window exit
    :param image: image path
    :return: mean of v channel
    """
    img = cv2.imread('capture/'+image)
    img_hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
    h, s, v = cv2.split(img_hsv)
    return numpy.mean(v)


def get_rectangle_contours(binary):
    _, contours, _ = cv2.findContours(binary, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    rectangle_contours = []
    for counter in contours:
        x, y, w, h = cv2.boundingRect(counter)
        cnt = numpy.array([[x, y], [x + w, y], [x + w, y + h], [x, y + h]])
        rectangle_contours.append(cnt)
    rectangle_contours = sorted(rectangle_contours, key=cv2.contourArea, reverse=True)[:100]
    rectangle_contours = contour_area_filter(binary, rectangle_contours)
    rectangle_contours = merge_rectangle_contours(rectangle_contours)
    return rectangle_contours


def get_center_pos(contour):
    x = int((contour[0][0]+contour[1][0])/2)
    y = int((contour[1][1]+contour[2][1])/2)
    return [x, y]


def get_label_pos(contour):
    center = get_center_pos(contour)
    x = int((int((center[0]+contour[2][0])/2)+contour[2][0])/2)
    y = int((int((center[1]+contour[2][1])/2)+contour[2][1])/2)
    return [x, y]


def draw_contours(img, contours, color="info"):
    if color == "info":
        cv2.drawContours(img, contours, -1, (255, 145, 30), 3)


def yolox_preprocess(img, input_size, swap=(2, 0, 1)):
    if len(img.shape) == 3:
        padded_img = numpy.ones((input_size[0], input_size[1], 3), dtype=numpy.uint8) * 114
    else:
        padded_img = numpy.ones(input_size, dtype=numpy.uint8) * 114
    r = min(input_size[0] / img.shape[0], input_size[1] / img.shape[1])
    resized_img = cv2.resize(
        img,
        (int(img.shape[1] * r), int(img.shape[0] * r)),
        interpolation=cv2.INTER_LINEAR,
    ).astype(numpy.uint8)
    padded_img[: int(img.shape[0] * r), : int(img.shape[1] * r)] = resized_img
    padded_img = padded_img.transpose(swap)
    padded_img = numpy.ascontiguousarray(padded_img, dtype=numpy.float32)
    return padded_img, r


def yolox_postprocess(outputs, img_size, p6=False):
    grids = []
    expanded_strides = []
    if not p6:
        strides = [8, 16, 32]
    else:
        strides = [8, 16, 32, 64]
    hsizes = [img_size[0] // stride for stride in strides]
    wsizes = [img_size[1] // stride for stride in strides]
    for hsize, wsize, stride in zip(hsizes, wsizes, strides):
        xv, yv = numpy.meshgrid(numpy.arange(wsize), numpy.arange(hsize))
        grid = numpy.stack((xv, yv), 2).reshape(1, -1, 2)
        grids.append(grid)
        shape = grid.shape[:2]
        expanded_strides.append(numpy.full((*shape, 1), stride))
    grids = numpy.concatenate(grids, 1)
    expanded_strides = numpy.concatenate(expanded_strides, 1)
    outputs[..., :2] = (outputs[..., :2] + grids) * expanded_strides
    outputs[..., 2:4] = numpy.exp(outputs[..., 2:4]) * expanded_strides
    return outputs


def nms(boxes, scores, nms_thr):
    """Single class NMS implemented in Numpy."""
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = numpy.maximum(x1[i], x1[order[1:]])
        yy1 = numpy.maximum(y1[i], y1[order[1:]])
        xx2 = numpy.minimum(x2[i], x2[order[1:]])
        yy2 = numpy.minimum(y2[i], y2[order[1:]])

        w = numpy.maximum(0.0, xx2 - xx1 + 1)
        h = numpy.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = numpy.where(ovr <= nms_thr)[0]
        order = order[inds + 1]

    return keep


def multiclass_nms(boxes, scores, nms_thr, score_thr, class_agnostic=True):
    """Multiclass NMS implemented in Numpy"""
    if class_agnostic:
        nms_method = multiclass_nms_class_agnostic
    else:
        nms_method = multiclass_nms_class_aware
    return nms_method(boxes, scores, nms_thr, score_thr)


def multiclass_nms_class_agnostic(boxes, scores, nms_thr, score_thr):
    """Multiclass NMS implemented in Numpy. Class-agnostic version."""
    cls_inds = scores.argmax(1)
    cls_scores = scores[numpy.arange(len(cls_inds)), cls_inds]

    valid_score_mask = cls_scores > score_thr
    if valid_score_mask.sum() == 0:
        return None
    valid_scores = cls_scores[valid_score_mask]
    valid_boxes = boxes[valid_score_mask]
    valid_cls_inds = cls_inds[valid_score_mask]
    keep = nms(valid_boxes, valid_scores, nms_thr)
    if keep:
        dets = numpy.concatenate(
            [valid_boxes[keep], valid_scores[keep, None], valid_cls_inds[keep, None]], 1
        )
    return dets


def multiclass_nms_class_aware(boxes, scores, nms_thr, score_thr):
    """Multiclass NMS implemented in Numpy. Class-aware version."""
    final_dets = []
    num_classes = scores.shape[1]
    for cls_ind in range(num_classes):
        cls_scores = scores[:, cls_ind]
        valid_score_mask = cls_scores > score_thr
        if valid_score_mask.sum() == 0:
            continue
        else:
            valid_scores = cls_scores[valid_score_mask]
            valid_boxes = boxes[valid_score_mask]
            keep = nms(valid_boxes, valid_scores, nms_thr)
            if len(keep) > 0:
                cls_inds = numpy.ones((len(keep), 1)) * cls_ind
                dets = numpy.concatenate(
                    [valid_boxes[keep], valid_scores[keep, None], cls_inds], 1
                )
                final_dets.append(dets)
    if len(final_dets) == 0:
        return None
    return numpy.concatenate(final_dets, 0)


def img_show(img, boxes, scores, cls_ids, conf=0.5, class_names=None):
    _COLORS = numpy.array([255, 0, 0,
                           195, 123, 40,
                           110, 176, 23]).astype(numpy.float32).reshape(-1, 3)
    for i in range(len(boxes)):
        box = boxes[i]
        cls_id = int(cls_ids[i])
        score = scores[i]
        if score < conf:
            continue
        x0 = int(box[0])
        y0 = int(box[1])
        x1 = int(box[2])
        y1 = int(box[3])

        color = _COLORS[cls_id].astype(numpy.uint8).tolist()
        text = '{}:{:.1f}%'.format(class_names[cls_id], score * 100)
        txt_color = (0, 0, 0) if numpy.mean(_COLORS[cls_id]) > 128 else (255, 255, 255)
        font = cv2.FONT_HERSHEY_SIMPLEX

        txt_size = cv2.getTextSize(text, font, 0.4, 1)[0]
        cv2.rectangle(img, (x0, y0), (x1, y1), color, 3)

        txt_bk_color = (_COLORS[cls_id] * 0.7).astype(numpy.uint8).tolist()
        cv2.rectangle(
            img,
            (x0, y0 + 1),
            (x0 + txt_size[0] + 1, y0 + int(1.5*txt_size[1])),
            txt_bk_color,
            -1
        )
        cv2.putText(img, text, (x0, y0 + txt_size[1]), font, 0.4, txt_color, thickness=1)

    return img


def download_image(img_url, image_name):
    if not os.path.exists(IMAGE_TEMP_DIR):
        os.makedirs(IMAGE_TEMP_DIR)
    image_path = os.path.join(IMAGE_TEMP_DIR, image_name)
    r = requests.get(img_url, stream=True)
    if r.status_code == 200:
        with open(image_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=32):
                f.write(chunk)
    else:
        raise Exception(f'Download image fail, image url reponse code: {r.status_code}')
    return image_path


def save_base64_image(base64_image, image_name):
    if not os.path.exists(IMAGE_TEMP_DIR):
        os.makedirs(IMAGE_TEMP_DIR)
    image_path = os.path.join(IMAGE_TEMP_DIR, image_name)
    img_str = base64.b64decode(base64_image)
    img_np_arr = numpy.fromstring(img_str, numpy.uint8)
    image = cv2.imdecode(img_np_arr, cv2.IMREAD_COLOR)
    cv2.imwrite(image_path, image)
    return image_path


def get_hash_score(hash1, hash2, precision=8):
    """
    calculate similar score with line A and line B
    :param hash1: input line A with hash code
    :param hash2: input line B with hash code
    :return: similar score in 0-1.0
    """
    assert len(hash1) == len(hash2)
    score = 1 - sum([ch1 != ch2 for ch1, ch2 in zip(hash1, hash2)]) * 1.0 / (precision * precision)
    return score


def m_diff(e, f, i=0, j=0, equal_obj=object()):
    N, M, L, Z = len(e), len(f), len(e) + len(f), 2 * min(len(e), len(f)) + 2
    if N > 0 and M > 0:
        w, g, p = N - M, [0] * Z, [0] * Z
        for h in range(0, (L // 2 + (L % 2 != 0)) + 1):
            for r in range(0, 2):
                c, d, o, m = (g, p, 1, 1) if r == 0 else (p, g, 0, -1)
                for k in range(-(h - 2 * max(0, h - M)), h - 2 * max(0, h - N) + 1, 2):
                    a = c[(k + 1) % Z] if (k == -h or k != h and c[(k - 1) % Z] < c[(k + 1) % Z]) else c[(k - 1) % Z] + 1
                    b = a - k
                    s, t = a, b
                    while a < N and b < M and equal_obj.equal(e[(1 - o) * N + m * a + (o - 1)], f[(1 - o) * M + m * b + (o - 1)]):
                        a, b = a + 1, b + 1
                    c[k % Z], z = a, -(k - w)
                    if L % 2 == o and -(h - o) <= z <= h - o and c[k % Z] + d[z % Z] >= N:
                        D, x, y, u, v = (2 * h - 1, s, t, a, b) if o == 1 else (2 * h, N - a, M - b, N - s, M - t)
                        if D > 1 or (x != u and y != v):
                            return m_diff(e[0:x], f[0:y], i, j, equal_obj) + m_diff(e[u:N], f[v:M], i + u, j + v, equal_obj)
                        elif M > N:
                            return m_diff([], f[N:M], i + N, j + N, equal_obj)
                        elif M < N:
                            return m_diff(e[M:N], [], i + M, j + M, equal_obj)
                        else:
                            return []
    elif N > 0:
        return [{"operation": "delete", "position_old": i + n} for n in range(0, N)]
    else:
        return [{"operation": "insert", "position_old": i, "position_new": j + n} for n in range(0, M)]


def compute_iou(rect1, rect2):
    """
    computing IoU
    :param rec1: (y0, x0, y1, x1), which reflects
            (top, left, bottom, right)
    :param rec2: (y0, x0, y1, x1)
    :return: scala value of IoU
    """
    rec1 = [rect1[1], rect1[0], rect1[3], rect1[2]]
    rec2 = [rect2[1], rect2[0], rect2[3], rect2[2]]
    # computing area of each rectangles
    S_rec1 = (rec1[2] - rec1[0]) * (rec1[3] - rec1[1])
    S_rec2 = (rec2[2] - rec2[0]) * (rec2[3] - rec2[1])

    # computing the sum_area
    sum_area = S_rec1 + S_rec2

    # find the each edge of intersect rectangle
    left_line = max(rec1[1], rec2[1])
    right_line = min(rec1[3], rec2[3])
    top_line = max(rec1[0], rec2[0])
    bottom_line = min(rec1[2], rec2[2])

    # judge if there is an intersect
    if left_line >= right_line or top_line >= bottom_line:
        return 0
    else:
        intersect = (right_line - left_line) * (bottom_line - top_line)
        return (intersect / (sum_area - intersect)) * 1.0


def proposal_fine_tune(scores, proposals, thresh):
    _ids = [i for i, score in enumerate(scores) if score > thresh]
    top_k_ids = numpy.argsort(scores)[-2:]
    if len(_ids) > 1:
        index_i = top_k_ids[0]
        index_j = top_k_ids[1]
        region_i = proposals[index_i]['elem_det_region']
        region_j = proposals[index_j]['elem_det_region']
        iou = compute_iou(region_i, region_j)
        if iou > 0.3:
            _x = int(((region_j[0] - region_i[0]) / 2))
            _y = int(((region_j[1] - region_i[1]) / 2))
            proposals[index_j]['elem_det_region'] = [region_j[0] - _x, region_j[1] - _y,
                                                     region_j[2] - _x, region_j[3] - _y]


def get_image_patches(img: numpy.ndarray, patch_w, patch_h, resolution_w: float, resolution_h: float):
    """
    滑动窗口生成patches
    """
    patches = []
    origin_h, origin_w, _ = img.shape
    assert patch_w <= origin_w and patch_h <= origin_h
    padding_w = numpy.zeros((origin_h, origin_w % patch_w, 3), numpy.uint8) + 255
    padding_h = numpy.zeros((origin_h % patch_h, origin_w + origin_w % patch_w, 3), numpy.uint8) + 255
    img_padding = numpy.vstack((numpy.hstack((img, padding_w)), padding_h))
    h, w, _ = img_padding.shape
    step_w = int(w / patch_w)
    step_h = int(h / patch_h)
    for i in numpy.arange(0, step_w, resolution_w):
        for j in numpy.arange(0, step_h, resolution_h):
            x0 = int(i*patch_w)
            y0 = int(j*patch_h)
            x1 = int((i+1)*patch_w) if (i+1)*patch_w < origin_w else origin_w
            y1 = int((j+1)*patch_h) if (j+1)*patch_h < origin_h else origin_h
            if x1 > x0 and y1 > y0:
                patches.append({'elem_det_region': [x0, y0, x1, y1]})
    return patches


def get_patches_value(img: numpy.ndarray, num_split: int):
    origin_h, origin_w, _ = img.shape
    padding_h = origin_h % num_split
    padding_w = origin_w % num_split
    _img = img[:origin_h-padding_h, :origin_w-padding_w, :]
    _h, _w, _ = _img.shape
    patches = get_image_patches(_img, int(_w/num_split), int(_h/num_split), 1.0, 1.0)
    ret = {}
    for i, patch in enumerate(patches):
        x1, y1, x2, y2 = list(map(int, patch['elem_det_region']))
        roi = get_roi_image(img, [[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
        ret[i] = int(numpy.mean(roi))
    return ret



import numpy as np
import string


class CharacterOps(object):
    """ Convert between text-label and text-index """

    def __init__(self, config):
        self.character_type = config['character_type']
        self.loss_type = config['loss_type']
        self.max_text_len = config['max_text_length']
        if self.character_type == "en":
            self.character_str = "0123456789abcdefghijklmnopqrstuvwxyz"
            dict_character = list(self.character_str)
        elif self.character_type == "ch":
            character_dict_path = config['character_dict_path']
            add_space = False
            if 'use_space_char' in config:
                add_space = config['use_space_char']
            self.character_str = ""
            with open(character_dict_path, "rb") as fin:
                lines = fin.readlines()
                for line in lines:
                    line = line.decode('utf-8').strip("\n").strip("\r\n")
                    self.character_str += line
            if add_space:
                self.character_str += " "
            dict_character = list(self.character_str)
        elif self.character_type == "en_sensitive":
            # same with ASTER setting (use 94 char).
            self.character_str = string.printable[:-6]
            dict_character = list(self.character_str)
        else:
            self.character_str = None
        assert self.character_str is not None, \
            "Nonsupport type of the character: {}".format(self.character_str)
        self.beg_str = "sos"
        self.end_str = "eos"
        if self.loss_type == "attention":
            dict_character = [self.beg_str, self.end_str] + dict_character
        elif self.loss_type == "srn":
            dict_character = dict_character + [self.beg_str, self.end_str]
        self.dict = {}
        for i, char in enumerate(dict_character):
            self.dict[char] = i
        self.character = dict_character

    def encode(self, text):
        """convert text-label into text-index.
        input:
            text: text labels of each image. [batch_size]

        output:
            text: concatenated text index for CTCLoss.
                    [sum(text_lengths)] = [text_index_0 + text_index_1 + ... + text_index_(n - 1)]
            length: length of each text. [batch_size]
        """
        if self.character_type == "en":
            text = text.lower()

        text_list = []
        for char in text:
            if char not in self.dict:
                continue
            text_list.append(self.dict[char])
        text = np.array(text_list)
        return text

    def decode(self, text_index, is_remove_duplicate=False):
        """ convert text-index into text-label. """
        char_list = []
        char_num = self.get_char_num()

        if self.loss_type == "attention":
            beg_idx = self.get_beg_end_flag_idx("beg")
            end_idx = self.get_beg_end_flag_idx("end")
            ignored_tokens = [beg_idx, end_idx]
        else:
            ignored_tokens = [char_num]

        for idx in range(len(text_index)):
            if text_index[idx] in ignored_tokens:
                continue
            if is_remove_duplicate:
                if idx > 0 and text_index[idx - 1] == text_index[idx]:
                    continue
            char_list.append(self.character[int(text_index[idx])])
        text = ''.join(char_list)
        return text

    def get_char_num(self):
        return len(self.character)

    def get_beg_end_flag_idx(self, beg_or_end):
        if self.loss_type == "attention":
            if beg_or_end == "beg":
                idx = np.array(self.dict[self.beg_str])
            elif beg_or_end == "end":
                idx = np.array(self.dict[self.end_str])
            else:
                assert False, "Unsupport type %s in get_beg_end_flag_idx"\
                    % beg_or_end
            return idx
        else:
            err = "error in get_beg_end_flag_idx when using the loss %s"\
                % (self.loss_type)
            assert False, err



import time

import cv2
import os

import numpy
import torch
import numpy as np
from PIL import Image
from scipy import spatial
from visionx.config import CLIP_MODEL_PATH, OP_NUM_THREADS
from visionx.service.image_infer import get_ui_infer
from visionx.dbnet_crnn.image_text import ImageText
from visionx.service.image_utils import get_roi_image, img_show, get_image_patches, proposal_fine_tune, get_infer_area
from torchvision.transforms import Compose, Resize, CenterCrop, ToTensor, Normalize
import onnxruntime

try:
    import clip
except ImportError:
    print('pip install git+https://github.com/openai/CLIP.git')


try:
    from torchvision.transforms import InterpolationMode
    BICUBIC = InterpolationMode.BICUBIC
except ImportError:
    BICUBIC = Image.BICUBIC


def cosine_similar(l1, l2):
    return 1 - spatial.distance.cosine(l1, l2)


def _convert_image_to_rgb(image):
    return image.convert("RGB")


def target_roi_text_diff_rate(target_img, source_img, proposals, tp):
    image_text = ImageText()
    count = 0
    target_text = image_text.get_text(target_img, target_img.shape[0])
    x1, y1, x2, y2 = list(map(int, proposals[tp[-1]]['elem_det_region']))
    roi = get_roi_image(source_img, [[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
    source_text = image_text.get_text(roi, roi.shape[0])
    if len(tp) > 1:
        x1, y1, x2, y2 = list(map(int, proposals[tp[-2]]['elem_det_region']))
        roi = get_roi_image(source_img, [[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
        source_text.extend(image_text.get_text(roi, roi.shape[0]))
    for t in target_text:
        for s in source_text:
            if t['text'] in s['text'] or s['text'] in t['text']:
                count = count + 1
                break
    rate = count / len(target_text) if len(target_text) > 0 else 0
    return rate


def get_proposals(target_image, source_image_path, provider="ui-infer", patches_resolution="normal"):
    """
    选择区域来源，只需提供位置
    """
    # ui-infer，模型推理
    if provider == 'ui-infer':
        image_infer_result = get_ui_infer(source_image_path, 0.01)
    # patches，滑动窗口
    else:
        h, w, _ = target_image.shape
        resolution_map = {
            'normal': [0.6, 0.6],
            'high': [0.3, 0.3]
        }
        resolution = resolution_map[patches_resolution]
        source_img = cv2.imread(source_image_path)
        _h, _w, _ = source_img.shape
        if w >= _w:
            ratio = _w / w
            target_image = cv2.resize(target_image, (0, 0), fx=ratio, fy=ratio)
            h, w, _ = target_image.shape
        resolution[0] = round(resolution[0]/2, 1) if _w / w < 6 else resolution[0]
        resolution[1] = round(resolution[1]/2, 1) if _h / h < 6 else resolution[1]
        image_infer_result = get_image_patches(source_img, w, h, resolution[0], resolution[1])
    return image_infer_result


class ImageTrace(object):
    def __init__(self):
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        print(f"Using {self.device}.\nStart loading model")
        self.n_px = 224
        self.template_target_image = np.zeros([100, 100, 3], dtype=np.uint8) + 100
        self.preprocess = self._get_preprocess()
        so = onnxruntime.SessionOptions()
        so.intra_op_num_threads = OP_NUM_THREADS
        self.ort_sess = onnxruntime.InferenceSession(CLIP_MODEL_PATH, sess_options=so)
        print("Finish loading")

    def _get_preprocess(self):
        return Compose([
            Resize(self.n_px, interpolation=BICUBIC),
            CenterCrop(self.n_px),
            _convert_image_to_rgb,
            ToTensor(),
            Normalize((0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)),
        ])

    def search_image(self, target_image_info: dict, source_image_path, top_k, image_alpha, text_alpha, provider):
        top_k = top_k  # 最大匹配数量
        image_alpha = image_alpha  # 图相关系数
        text_alpha = text_alpha  # 文本相关系数
        roi_list = []
        img_text_score = []
        target_image = target_image_info.get('img', self.template_target_image)
        target_image = cv2.imread(target_image) if isinstance(target_image, str) else target_image
        target_image_desc = target_image_info.get('desc', '')
        source_image = cv2.imread(source_image_path)
        proposals = get_proposals(target_image, source_image_path, provider=provider)
        text = clip.tokenize([target_image_desc]).to(self.device)
        # 提取检测目标
        for roi in proposals:
            x1, y1, x2, y2 = list(map(int, roi['elem_det_region']))
            roi = get_roi_image(source_image, [[x1, y1], [x2, y1], [x2, y2], [x1, y2]])
            img_pil = Image.fromarray(cv2.cvtColor(roi, cv2.COLOR_BGR2RGB))
            roi_list.append(self.preprocess(img_pil).to(self.device))
        # 计算图像和文本匹配向量
        img_pil = Image.fromarray(cv2.cvtColor(target_image, cv2.COLOR_BGR2RGB))
        target_image_input = self.preprocess(img_pil).unsqueeze(0).to(self.device).clone().detach()
        source_image_input = torch.tensor(np.stack(roi_list))
        _, logits_per_text,  source_image_features, = self.ort_sess.run(
            ["LOGITS_PER_IMAGE", "LOGITS_PER_TEXT", "onnx::Mul_3493"],
            {"IMAGE": source_image_input.numpy(), "TEXT": text.numpy()}
        )
        probs = torch.from_numpy(logits_per_text).softmax(dim=-1).cpu().numpy()
        if image_alpha != 0:
            target_image_features = self.ort_sess.run(
                ["onnx::Mul_3493"], {"IMAGE": target_image_input.numpy(), "TEXT": text.numpy()}
            )
        else:
            target_image_features = numpy.zeros([len(target_image_input)], dtype=np.uint8)
        # 图像加文本
        for i, source_image_feature in enumerate(source_image_features):
            score = cosine_similar(target_image_features[0][0], source_image_feature) if image_alpha != 0 else 0
            img_text_score.append(score * image_alpha + probs[0][i] * text_alpha)
        max_confidence = round(np.max(img_text_score) / (image_alpha + text_alpha), 3)
        target_image_infer = get_ui_infer(target_image, 0.1)
        target_area_rate = get_infer_area(target_image_infer) / (target_image.shape[0] * target_image.shape[1])
        score_norm = (img_text_score - np.min(img_text_score)) / (np.max(img_text_score) - np.min(img_text_score))
        top_k_ids = np.argsort(score_norm)[-top_k:]
        proposal_fine_tune(score_norm, proposals, 0.9)
        # CLIP refer text in image
        if target_area_rate < 0.1 and provider == 'patches':
            text_rate = target_roi_text_diff_rate(target_image, source_image, proposals, np.argsort(score_norm)[:])
            max_confidence = max_confidence - 0.06*((0.1-target_area_rate)/0.1) + 0.1*text_rate
        return top_k_ids, score_norm, proposals, max_confidence

    def get_trace_result(self, target_image_info, source_image_path, top_k=3, image_alpha=1.0,
                         text_alpha=0.6, proposal_provider='ui-infer'):
        top_k_ids, scores, infer_result, max_confidence = self.search_image(
            target_image_info, source_image_path, top_k, image_alpha, text_alpha, proposal_provider
        )
        print(f"Max confidence:{max_confidence}")
        cls_ids = np.zeros(len(top_k_ids), dtype=int)
        boxes = [infer_result[i]['elem_det_region'] for i in top_k_ids]
        scores = [float(scores[i]) for i in top_k_ids]
        image_show = img_show(cv2.imread(source_image_path), boxes, scores, cls_ids, conf=0.5, class_names=['T'])
        return image_show

    def video_target_track(self, video_path, target_image_info, work_path):
        video_cap = cv2.VideoCapture(video_path)
        _, im = video_cap.read()
        fourcc = cv2.VideoWriter_fourcc(*"mp4v")
        im_save_path = os.path.join(work_path, 'im_temp.png')
        video_out_path = os.path.join(work_path, 'video_out.mp4')
        out = cv2.VideoWriter(video_out_path, fourcc, 20, (im.shape[1], im.shape[0]))
        i = 0
        while 1:
            i = i + 1
            if i % 2 == 0:
                continue
            print(f"video parsing {i}")
            ret, im = video_cap.read()
            if ret:
                cv2.imwrite(im_save_path, im)
                trace_result = self.get_trace_result(target_image_info, im_save_path, top_k=1)
                out.write(trace_result)
            else:
                print("finish.")
                out.release()
                break


image_trace = ImageTrace()


def trace_target_video():
    target_image_info = {
        'path': "./capture/local_images/img_play_icon.png",
        'desc': "picture with play button"
    }
    target_image_info['img'] = cv2.imread(target_image_info['path'])
    video_path = "./capture/local_images/video.mp4"
    work_path = './capture/local_images'
    image_trace.video_target_track(video_path, target_image_info, work_path)


def search_target_image():
    """
    # robust target image search
    """
    # 图像目标系数
    image_alpha = 0.0
    # 文本描述系数
    text_alpha = 1.0
    # 最大匹配目标数量
    top_k = 1
    # 调试用，构造目标图像
    target_img = np.zeros([100, 100, 3], dtype=np.uint8)+255
    cv2.putText(target_img, 'Q', (30, 60), cv2.FONT_HERSHEY_SIMPLEX, 2.0, (0, 0, 0), thickness=3)
    # target_img = cv2.imread("./capture/local_images/mario.png")
    # 目标语言描述
    desc = "mario"
    target_image_info = {'img': target_img, 'desc': desc}
    source_image_path = "./capture/image_2.png"
    trace_result_path = "./capture/local_images/"
    if not os.path.exists(trace_result_path):
        os.mkdir(trace_result_path)
    # 查找目标
    t1 = time.time()
    image_trace_show = image_trace.get_trace_result(target_image_info, source_image_path, top_k=top_k,
                                                    image_alpha=image_alpha, text_alpha=text_alpha,
                                                    proposal_provider='ui-infer')
    print(f"Infer time:{round(time.time() - t1, 3)} s", )
    cv2.imwrite(trace_result_path+'trace_result.png', image_trace_show)
    print(f"Result saved {trace_result_path}")


if __name__ == '__main__':
    assert os.path.exists(CLIP_MODEL_PATH)
    search_target_image()



import cv2
from visionx.service.image_similar import HashSimilar
from visionx.service.image_utils import get_hash_score, m_diff


class LineFeatureEqual(object):
    def __init__(self):
        self.thresh = 0.85

    def equal(self, a, b):
        if get_hash_score(a, b) > self.thresh:
            return True
        else:
            return False


class ImageDiff(object):
    def __init__(self, w=9, padding=80, w_scale=850, h_scale=0.08, pixel_value=28):
        self.filter_w = w
        self.padding = padding
        self.size_scale = w_scale
        self.head_scale = h_scale
        self.pixel_value = pixel_value

    @staticmethod
    def get_line_list(op_list):
        """
        get line list
        :param op_list: op list
        :return: line list
        """
        line1_list = []
        line2_list = []
        for op in op_list:
            if op["operation"] == "insert":
                line1_list.append(op["position_new"])
            if op["operation"] == "delete":
                line2_list.append(op["position_old"])
        return line1_list, line2_list

    @staticmethod
    def get_line_feature(image, precision=8):
        """
        get line feature of input image
        :param image: image in numpy shape
        :param precision: feature precision
        :return: line feature
        """
        line_feature = []
        for y in range(image.shape[0]):
            img = cv2.resize(image[y], (precision, precision))
            img_list = img.flatten()
            avg = sum(img_list) * 1. / len(img_list)
            avg_list = ["0" if i < avg else "1" for i in img_list]
            line_feature.append([int(''.join(avg_list[x:x + 4]), 2) for x in range(0, precision * precision)])
        return line_feature

    def get_image_feature(self, img1, img2):
        """
        get image feature with padding processing
        :param img1: imageA in numpy shape
        :param img2: imageB in numpy shape
        :return: image feature
        """
        h1, w = img1.shape
        img1 = img1[:, :w - self.padding]
        img2 = img2[:, :w - self.padding]
        img1_feature = self.get_line_feature(img1)
        img2_feature = self.get_line_feature(img2)
        return img1_feature, img2_feature

    def line_filter(self, line_list):
        """
        calculate line list with param
        :param line_list: line list
        :return: filtered line list
        """
        i = 0
        w = self.filter_w
        line = []
        while i < len(line_list) - w - 1:
            f = line_list[i:i + w]
            s = 0
            for j in range(w - 1):
                s = s + f[j + 1] - f[j]
            if s - w <= 6:
                for l in f:
                    if l not in line:
                        line.append(l)
            i = i + 1
        return line

    def get_image(self, image_file):
        """
        cv2.read image and 3d to 1d
        :param image_file: image file path
        :return: image in numpy shape
        """
        image = cv2.imread(image_file)
        img = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        img = cv2.GaussianBlur(img, (5, 5), 1.5)
        h, w = img.shape
        scale = self.size_scale / w
        img = cv2.resize(img, (0, 0), fx=scale, fy=scale)
        return img

    @staticmethod
    def get_pixel(img, x, y):
        """
        get pixel value of image
        :param img: image in numpy shape
        :param x: position x
        :param y: position y
        :return: pixel value
        """
        h, w = img.shape
        p = 0
        if y < h:
            p = img[y][x]
        return p

    def increment_diff(self, image1, image2, image_show) -> int:
        """
        calculate increment image diff
        :param image1: input image A
        :param image2: input image B
        :param image_show: increment diff image for show
        :return: points length of image show
        """
        img1 = self.get_image(image1)
        img2 = self.get_image(image2)
        score_list = HashSimilar.get_attention(img1, img2)
        img1_feature, img2_feature = self.get_image_feature(img1, img2)
        line1, line2 = self.get_line_list(m_diff(img1_feature, img2_feature, equal_obj=LineFeatureEqual()))
        line = line1 + line2
        line = self.line_filter(line)
        img_show = img2.copy() if img2.shape[0] > img1.shape[0] else img1.copy()
        (h, w) = img_show.shape
        img_show = cv2.cvtColor(img_show, cv2.COLOR_GRAY2BGR)
        points = []
        line_attention = []
        for l in line:
            i = int((len(score_list) * (l - 1) / h))
            i = 0 if i < 0 else i
            if score_list[i] < 0.98:
                line_attention.append(l)
        line = line_attention
        for y in range(int(h * 0.95)):
            if y > int(w * self.head_scale):
                if y in line:
                    for x in range(w - self.padding):
                        p1 = int(self.get_pixel(img1, x, y))
                        p2 = int(self.get_pixel(img2, x, y))
                        if abs(p1 - p2) < self.pixel_value:
                            pass
                        else:
                            points.append([x, y])
        for point in points:
            cv2.circle(img_show, (point[0], point[1]), 1, (0, 0, 255), -1)
        cv2.imwrite(image_show, img_show)
        return len(points)

    def get_image_score(self, image1, image2, image_diff_name):
        """
        0.2-页面不一致，不生成对比图
        0.8-增量对比，生成对应路径的对比图
        1.0-页面相同，没有对比图
        :param image1:
        :param image2:
        :param image_diff_name:
        :return:
        """
        score = HashSimilar.get_attention_similar(image1, image2)
        if score < 1.0:
            if score > 0.2:
                points_size = self.increment_diff(image1, image2, image_diff_name)
                if points_size < 50:
                    score = 1.0
        return score



import cv2
import numpy as np
import math
import paddle.fluid as fluid
import visionx.dbnet_crnn.tools.utility as utility
from visionx.dbnet_crnn.ppocr.utils.character import CharacterOps


class TextRecognizer(object):
    def __init__(self, args, model_path):
        self.predictor, self.input_tensor, self.output_tensors =\
            utility.create_predictor(args, mode="rec", model_path=model_path)
        self.rec_image_shape = [int(v) for v in args['rec_image_shape'].split(",")]
        self.character_type = args['rec_char_type']
        self.rec_batch_num = args['rec_batch_num']
        self.rec_algorithm = args['rec_algorithm']
        self.text_len = args['max_text_length']
        self.use_zero_copy_run = args['use_zero_copy_run']
        char_ops_params = {
            "character_type": args['rec_char_type'],
            "character_dict_path": args['rec_char_dict_path'],
            "use_space_char": args['use_space_char'],
            "max_text_length": args['max_text_length']
        }
        if self.rec_algorithm in ["CRNN", "Rosetta", "STAR-Net"]:
            char_ops_params['loss_type'] = 'ctc'
            self.loss_type = 'ctc'
        self.char_ops = CharacterOps(char_ops_params)

    def resize_norm_img(self, img, max_wh_ratio):
        imgC, imgH, imgW = self.rec_image_shape
        assert imgC == img.shape[2]
        if self.character_type == "ch":
            imgW = int((32 * max_wh_ratio))
        h, w = img.shape[:2]
        ratio = w / float(h)
        if math.ceil(imgH * ratio) > imgW:
            resized_w = imgW
        else:
            resized_w = int(math.ceil(imgH * ratio))
        resized_image = cv2.resize(img, (resized_w, imgH))
        resized_image = resized_image.astype('float32')
        resized_image = resized_image.transpose((2, 0, 1)) / 255
        resized_image -= 0.5
        resized_image /= 0.5
        padding_im = np.zeros((imgC, imgH, imgW), dtype=np.float32)
        padding_im[:, :, 0:resized_w] = resized_image
        return padding_im

    def resize_norm_img_srn(self, img, image_shape):
        imgC, imgH, imgW = image_shape
        img_black = np.zeros((imgH, imgW))
        im_hei = img.shape[0]
        im_wid = img.shape[1]
        if im_wid <= im_hei * 1:
            img_new = cv2.resize(img, (imgH * 1, imgH))
        elif im_wid <= im_hei * 2:
            img_new = cv2.resize(img, (imgH * 2, imgH))
        elif im_wid <= im_hei * 3:
            img_new = cv2.resize(img, (imgH * 3, imgH))
        else:
            img_new = cv2.resize(img, (imgW, imgH))
        img_np = np.asarray(img_new)
        img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2GRAY)
        img_black[:, 0:img_np.shape[1]] = img_np
        img_black = img_black[:, :, np.newaxis]
        row, col, c = img_black.shape
        c = 1
        return np.reshape(img_black, (c, row, col)).astype(np.float32)

    def srn_other_inputs(self, image_shape, num_heads, max_text_length,
                         char_num):

        imgC, imgH, imgW = image_shape
        feature_dim = int((imgH / 8) * (imgW / 8))

        encoder_word_pos = np.array(range(0, feature_dim)).reshape(
            (feature_dim, 1)).astype('int64')
        gsrm_word_pos = np.array(range(0, max_text_length)).reshape(
            (max_text_length, 1)).astype('int64')

        gsrm_attn_bias_data = np.ones((1, max_text_length, max_text_length))
        gsrm_slf_attn_bias1 = np.triu(gsrm_attn_bias_data, 1).reshape(
            [-1, 1, max_text_length, max_text_length])
        gsrm_slf_attn_bias1 = np.tile(
            gsrm_slf_attn_bias1,
            [1, num_heads, 1, 1]).astype('float32') * [-1e9]

        gsrm_slf_attn_bias2 = np.tril(gsrm_attn_bias_data, -1).reshape(
            [-1, 1, max_text_length, max_text_length])
        gsrm_slf_attn_bias2 = np.tile(
            gsrm_slf_attn_bias2,
            [1, num_heads, 1, 1]).astype('float32') * [-1e9]

        encoder_word_pos = encoder_word_pos[np.newaxis, :]
        gsrm_word_pos = gsrm_word_pos[np.newaxis, :]

        return [
            encoder_word_pos, gsrm_word_pos, gsrm_slf_attn_bias1,
            gsrm_slf_attn_bias2
        ]

    def process_image_srn(self,
                          img,
                          image_shape,
                          num_heads,
                          max_text_length,
                          char_ops=None):
        norm_img = self.resize_norm_img_srn(img, image_shape)
        norm_img = norm_img[np.newaxis, :]
        char_num = char_ops.get_char_num()

        [encoder_word_pos, gsrm_word_pos, gsrm_slf_attn_bias1, gsrm_slf_attn_bias2] = \
            self.srn_other_inputs(image_shape, num_heads, max_text_length, char_num)

        gsrm_slf_attn_bias1 = gsrm_slf_attn_bias1.astype(np.float32)
        gsrm_slf_attn_bias2 = gsrm_slf_attn_bias2.astype(np.float32)

        return (norm_img, encoder_word_pos, gsrm_word_pos, gsrm_slf_attn_bias1,
                gsrm_slf_attn_bias2)

    def __call__(self, img_list):
        img_num = len(img_list)
        # Calculate the aspect ratio of all text bars
        width_list = []
        for img in img_list:
            width_list.append(img.shape[1] / float(img.shape[0]))
        # Sorting can speed up the recognition process
        indices = np.argsort(np.array(width_list))
        rec_res = [['', 0.0]] * img_num
        batch_num = self.rec_batch_num
        for beg_img_no in range(0, img_num, batch_num):
            end_img_no = min(img_num, beg_img_no + batch_num)
            norm_img_batch = []
            max_wh_ratio = 0
            for ino in range(beg_img_no, end_img_no):
                h, w = img_list[indices[ino]].shape[0:2]
                wh_ratio = w * 1.0 / h
                max_wh_ratio = max(max_wh_ratio, wh_ratio)
            for ino in range(beg_img_no, end_img_no):
                if self.loss_type != "srn":
                    norm_img = self.resize_norm_img(img_list[indices[ino]],
                                                    max_wh_ratio)
                    norm_img = norm_img[np.newaxis, :]
                    norm_img_batch.append(norm_img)
                else:
                    norm_img = self.process_image_srn(img_list[indices[ino]],
                                                      self.rec_image_shape, 8,
                                                      25, self.char_ops)
                    encoder_word_pos_list = []
                    gsrm_word_pos_list = []
                    gsrm_slf_attn_bias1_list = []
                    gsrm_slf_attn_bias2_list = []
                    encoder_word_pos_list.append(norm_img[1])
                    gsrm_word_pos_list.append(norm_img[2])
                    gsrm_slf_attn_bias1_list.append(norm_img[3])
                    gsrm_slf_attn_bias2_list.append(norm_img[4])
                    norm_img_batch.append(norm_img[0])

            norm_img_batch = np.concatenate(norm_img_batch, axis=0)
            norm_img_batch = norm_img_batch.copy()

            if self.loss_type == "srn":
                encoder_word_pos_list = np.concatenate(encoder_word_pos_list)
                gsrm_word_pos_list = np.concatenate(gsrm_word_pos_list)
                gsrm_slf_attn_bias1_list = np.concatenate(
                    gsrm_slf_attn_bias1_list)
                gsrm_slf_attn_bias2_list = np.concatenate(
                    gsrm_slf_attn_bias2_list)
                norm_img_batch = fluid.core.PaddleTensor(norm_img_batch)
                encoder_word_pos_list = fluid.core.PaddleTensor(
                    encoder_word_pos_list)
                gsrm_word_pos_list = fluid.core.PaddleTensor(gsrm_word_pos_list)
                gsrm_slf_attn_bias1_list = fluid.core.PaddleTensor(
                    gsrm_slf_attn_bias1_list)
                gsrm_slf_attn_bias2_list = fluid.core.PaddleTensor(
                    gsrm_slf_attn_bias2_list)

                inputs = [
                    norm_img_batch, encoder_word_pos_list,
                    gsrm_slf_attn_bias1_list, gsrm_slf_attn_bias2_list,
                    gsrm_word_pos_list
                ]

                self.predictor.run(inputs)
            else:
                if self.use_zero_copy_run:
                    self.input_tensor.copy_from_cpu(norm_img_batch)
                    self.predictor.zero_copy_run()
                else:
                    norm_img_batch = fluid.core.PaddleTensor(norm_img_batch)
                    self.predictor.run([norm_img_batch])

            if self.loss_type == "ctc":
                rec_idx_batch = self.output_tensors[0].copy_to_cpu()
                rec_idx_lod = self.output_tensors[0].lod()[0]
                predict_batch = self.output_tensors[1].copy_to_cpu()
                predict_lod = self.output_tensors[1].lod()[0]
                for rno in range(len(rec_idx_lod) - 1):
                    beg = rec_idx_lod[rno]
                    end = rec_idx_lod[rno + 1]
                    rec_idx_tmp = rec_idx_batch[beg:end, 0]
                    preds_text = self.char_ops.decode(rec_idx_tmp)
                    beg = predict_lod[rno]
                    end = predict_lod[rno + 1]
                    probs = predict_batch[beg:end, :]
                    ind = np.argmax(probs, axis=1)
                    blank = probs.shape[1]
                    valid_ind = np.where(ind != (blank - 1))[0]
                    if len(valid_ind) == 0:
                        continue
                    score = np.mean(probs[valid_ind, ind[valid_ind]])
                    rec_res[indices[beg_img_no + rno]] = [preds_text, score]
            elif self.loss_type == 'srn':
                rec_idx_batch = self.output_tensors[0].copy_to_cpu()
                probs = self.output_tensors[1].copy_to_cpu()
                char_num = self.char_ops.get_char_num()
                preds = rec_idx_batch.reshape(-1)
                total_preds = preds.copy()
                for ino in range(int(len(rec_idx_batch) / self.text_len)):
                    preds = total_preds[ino * self.text_len:(ino + 1) *
                                        self.text_len]
                    ind = np.argmax(probs, axis=1)
                    valid_ind = np.where(preds != int(char_num - 1))[0]
                    if len(valid_ind) == 0:
                        continue
                    score = np.mean(probs[valid_ind, ind[valid_ind]])
                    preds = preds[:valid_ind[-1] + 1]
                    preds_text = self.char_ops.decode(preds)

                    rec_res[indices[beg_img_no + ino]] = [preds_text, score]
            else:
                rec_idx_batch = self.output_tensors[0].copy_to_cpu()
                predict_batch = self.output_tensors[1].copy_to_cpu()
                for rno in range(len(rec_idx_batch)):
                    end_pos = np.where(rec_idx_batch[rno, :] == 1)[0]
                    if len(end_pos) <= 1:
                        preds = rec_idx_batch[rno, 1:]
                        score = np.mean(predict_batch[rno, 1:])
                    else:
                        preds = rec_idx_batch[rno, 1:end_pos[1]]
                        score = np.mean(predict_batch[rno, 1:end_pos[1]])
                    preds_text = self.char_ops.decode(preds)
                    rec_res[indices[beg_img_no + rno]] = [preds_text, score]
        return rec_res



import cv2
import numpy as np
import sys


class DBProcessTest(object):
    """
    DB pre-process for Test mode
    """

    def __init__(self):
        super(DBProcessTest, self).__init__()
        self.resize_type = 0

    def resize_image_type0(self, im, max_side_len):
        """
        resize image to a size multiple of 32 which is required by the network
        args:
            img(array): array with shape [h, w, c]
        return(tuple):
            img, (ratio_h, ratio_w)
        """
        h, w, _ = im.shape

        resize_w = w
        resize_h = h

        # limit the max side
        if max(resize_h, resize_w) > max_side_len:
            if resize_h > resize_w:
                ratio = float(max_side_len) / resize_h
            else:
                ratio = float(max_side_len) / resize_w
        else:
            ratio = 1.
        resize_h = int(resize_h * ratio)
        resize_w = int(resize_w * ratio)
        if resize_h % 32 == 0:
            resize_h = resize_h
        elif resize_h // 32 <= 1:
            resize_h = 32
        else:
            resize_h = (resize_h // 32 - 1) * 32
        if resize_w % 32 == 0:
            resize_w = resize_w
        elif resize_w // 32 <= 1:
            resize_w = 32
        else:
            resize_w = (resize_w // 32 - 1) * 32
        try:
            if int(resize_w) <= 0 or int(resize_h) <= 0:
                return None, (None, None)
            im = cv2.resize(im, (int(resize_w), int(resize_h)))
        except Exception as e:
            print(im.shape, resize_w, resize_h)
            sys.exit(0)
        ratio_h = resize_h / float(h)
        ratio_w = resize_w / float(w)
        return im, (ratio_h, ratio_w)

    def resize_image_type1(self, im):
        resize_h, resize_w = self.image_shape
        ori_h, ori_w = im.shape[:2]  # (h, w, c)
        im = cv2.resize(im, (int(resize_w), int(resize_h)))
        ratio_h = float(resize_h) / ori_h
        ratio_w = float(resize_w) / ori_w
        return im, (ratio_h, ratio_w)

    def normalize(self, im):
        img_mean = [0.485, 0.456, 0.406]
        img_std = [0.229, 0.224, 0.225]
        im = im.astype(np.float32, copy=False)
        im = im / 255
        im[:, :, 0] -= img_mean[0]
        im[:, :, 1] -= img_mean[1]
        im[:, :, 2] -= img_mean[2]
        im[:, :, 0] /= img_std[0]
        im[:, :, 1] /= img_std[1]
        im[:, :, 2] /= img_std[2]
        channel_swap = (2, 0, 1)
        im = im.transpose(channel_swap)
        return im

    def __call__(self, im, max_side_len):
        if self.resize_type == 0:
            im, (ratio_h, ratio_w) = self.resize_image_type0(im, max_side_len)
        else:
            im, (ratio_h, ratio_w) = self.resize_image_type1(im, max_side_len)
        im = self.normalize(im)
        im = im[np.newaxis, :]
        return [im, (ratio_h, ratio_w)]



import cv2
import copy
import numpy as np
import visionx.dbnet_crnn.tools.utility as utility
from visionx.config import PREDICT_DET_PATH, PREDICT_REC_PATH
from visionx.service.image_utils import get_center_pos
import visionx.dbnet_crnn.tools.predict_det as predict_det
import visionx.dbnet_crnn.tools.predict_rec as predict_rec


def sorted_boxes(dt_boxes):
    """
    Sort text boxes in order from top to bottom, left to right
    args:
        dt_boxes(array):detected text boxes with shape [4, 2]
    return:
        sorted boxes(array) with shape [4, 2]
    """
    num_boxes = dt_boxes.shape[0]
    sorted_boxes = sorted(dt_boxes, key=lambda x: (x[0][1], x[0][0]))
    _boxes = list(sorted_boxes)

    for i in range(num_boxes - 1):
        if abs(_boxes[i+1][0][1] - _boxes[i][0][1]) < 10 and (_boxes[i + 1][0][0] < _boxes[i][0][0]):
            tmp = _boxes[i]
            _boxes[i] = _boxes[i + 1]
            _boxes[i + 1] = tmp
    return _boxes


class ImageText(object):
    def __init__(self):
        args = utility.parse_args()
        self.text_detector = predict_det.TextDetector(args, model_path=PREDICT_DET_PATH)
        self.text_recognizer = predict_rec.TextRecognizer(args, model_path=PREDICT_REC_PATH)

    def get_rotate_crop_image(self, img, points):
        '''
        img_height, img_width = img.shape[0:2]
        left = int(np.min(points[:, 0]))
        right = int(np.max(points[:, 0]))
        top = int(np.min(points[:, 1]))
        bottom = int(np.max(points[:, 1]))
        img_crop = img[top:bottom, left:right, :].copy()
        points[:, 0] = points[:, 0] - left
        points[:, 1] = points[:, 1] - top
        '''
        img_crop_width = int(
            max(
                np.linalg.norm(points[0] - points[1]),
                np.linalg.norm(points[2] - points[3])))
        img_crop_height = int(
            max(
                np.linalg.norm(points[0] - points[3]),
                np.linalg.norm(points[1] - points[2])))
        pts_std = np.float32([[0, 0], [img_crop_width, 0],
                              [img_crop_width, img_crop_height],
                              [0, img_crop_height]])
        M = cv2.getPerspectiveTransform(points, pts_std)
        dst_img = cv2.warpPerspective(img, M, (img_crop_width, img_crop_height),
                                      borderMode=cv2.BORDER_REPLICATE,
                                      flags=cv2.INTER_CUBIC)
        dst_img_height, dst_img_width = dst_img.shape[0:2]
        if dst_img_height * 1.0 / dst_img_width >= 1.5:
            dst_img = np.rot90(dst_img)
        return dst_img

    def get_ocr(self, img, max_side_len):
        ori_im = img.copy()
        dt_boxes = self.text_detector(img, max_side_len)
        if dt_boxes is None:
            return None, None
        img_crop_list = []
        dt_boxes = sorted_boxes(dt_boxes)
        for bno in range(len(dt_boxes)):
            tmp_box = copy.deepcopy(dt_boxes[bno])
            img_crop = self.get_rotate_crop_image(ori_im, tmp_box)
            img_crop_list.append(img_crop)
        rec_res = self.text_recognizer(img_crop_list)
        return dt_boxes, rec_res

    def get_text(self, img, max_side_len, score_thresh=0.6):
        result = []
        dt_boxes, rec_res = self.get_ocr(img, max_side_len)
        for roi_ocr in list(zip(dt_boxes, rec_res)):
            roi_score = roi_ocr[1][1]
            if roi_score > score_thresh:
                boxes = roi_ocr[0]
                result.append({
                    'pos': get_center_pos(roi_ocr[0]),
                    'text': roi_ocr[1][0],
                    'score': round(float(roi_score), 2),
                    'elem_det_region': [boxes[0][0], boxes[0][1], boxes[2][0], boxes[2][1]]
                })
        return result


image_text = ImageText()



import os
from paddle.fluid.core import AnalysisConfig
from paddle.fluid.core import create_paddle_predictor

from visionx.config import REC_CHAR_DICT_PATH


def parse_args():
    params = dict()

    # prediction engine
    params["r_optim"] = True
    params["use_tensorrt"] = False
    params["gpu_mem"] = 8000

    # text detector
    params["det_algorithm"] = 'DB'
    params["det_max_side_len"] = 1500

    # DB Net
    params["det_db_thresh"] = 0.3
    params["det_db_box_thresh"] = 0.5
    params["det_db_unclip_ratio"] = 2.0

    # text recognizer
    params["rec_algorithm"] = 'CRNN'
    params["rec_image_shape"] = "3, 32, 320"
    params["rec_char_type"] = 'ch'
    params["rec_batch_num"] = 30
    params["max_text_length"] = 25
    params["rec_char_dict_path"] = REC_CHAR_DICT_PATH
    params["use_space_char"] = True
    params["enable_mkldnn"] = False
    params["use_zero_copy_run"] = False
    return params


def create_predictor(args, mode, model_path):
    model_dir = model_path
    model_file_path = model_dir + "/model"
    params_file_path = model_dir + "/params"
    assert os.path.exists(model_file_path)
    assert os.path.exists(params_file_path)
    config = AnalysisConfig(model_file_path, params_file_path)

    # use CPU
    config.disable_gpu()
    config.set_cpu_math_library_num_threads(6)
    if args['enable_mkldnn']:
        config.enable_mkldnn()

    # config.enable_memory_optim()
    config.disable_glog_info()

    if args['use_zero_copy_run']:
        config.delete_pass("conv_transpose_eltwiseadd_bn_fuse_pass")
        config.switch_use_feed_fetch_ops(False)
    else:
        config.switch_use_feed_fetch_ops(True)

    predictor = create_paddle_predictor(config)
    input_names = predictor.get_input_names()
    for name in input_names:
        input_tensor = predictor.get_input_tensor(name)
    output_names = predictor.get_output_names()
    output_tensors = []
    for output_name in output_names:
        output_tensor = predictor.get_output_tensor(output_name)
        output_tensors.append(output_tensor)
    return predictor, input_tensor, output_tensors



import numpy as np
import cv2
from shapely.geometry import Polygon
import pyclipper


class DBPostProcess(object):
    """
    The post process for Differentiable Binarization (DB).
    """

    def __init__(self, params):
        self.thresh = params['thresh']
        self.box_thresh = params['box_thresh']
        self.max_candidates = params['max_candidates']
        self.unclip_ratio = params['unclip_ratio']
        self.min_size = 3

    def boxes_from_bitmap(self, pred, _bitmap, dest_width, dest_height):
        '''
        _bitmap: single map with shape (1, H, W),
                whose values are binarized as {0, 1}
        '''

        bitmap = _bitmap
        height, width = bitmap.shape

        outs = cv2.findContours((bitmap * 255).astype(np.uint8), cv2.RETR_LIST,
                                cv2.CHAIN_APPROX_SIMPLE)
        if len(outs) == 3:
            img, contours, _ = outs[0], outs[1], outs[2]
        elif len(outs) == 2:
            contours, _ = outs[0], outs[1]

        num_contours = min(len(contours), self.max_candidates)
        boxes = np.zeros((num_contours, 4, 2), dtype=np.int16)
        scores = np.zeros((num_contours, ), dtype=np.float32)

        for index in range(num_contours):
            contour = contours[index]
            points, sside = self.get_mini_boxes(contour)
            if sside < self.min_size:
                continue
            points = np.array(points)
            score = self.box_score_fast(pred, points.reshape(-1, 2))
            if self.box_thresh > score:
                continue

            box = self.unclip(points).reshape(-1, 1, 2)
            box, sside = self.get_mini_boxes(box)
            if sside < self.min_size + 2:
                continue
            box = np.array(box)
            if not isinstance(dest_width, int):
                dest_width = dest_width.item()
                dest_height = dest_height.item()

            box[:, 0] = np.clip(
                np.round(box[:, 0] / width * dest_width), 0, dest_width)
            box[:, 1] = np.clip(
                np.round(box[:, 1] / height * dest_height), 0, dest_height)
            boxes[index, :, :] = box.astype(np.int16)
            scores[index] = score
        return boxes, scores

    def unclip(self, box):
        unclip_ratio = self.unclip_ratio
        poly = Polygon(box)
        distance = poly.area * unclip_ratio / poly.length
        offset = pyclipper.PyclipperOffset()
        offset.AddPath(box, pyclipper.JT_ROUND, pyclipper.ET_CLOSEDPOLYGON)
        expanded = np.array(offset.Execute(distance))
        return expanded

    def get_mini_boxes(self, contour):
        bounding_box = cv2.minAreaRect(contour)
        points = sorted(list(cv2.boxPoints(bounding_box)), key=lambda x: x[0])
        if points[1][1] > points[0][1]:
            index_1 = 0
            index_4 = 1
        else:
            index_1 = 1
            index_4 = 0
        if points[3][1] > points[2][1]:
            index_2 = 2
            index_3 = 3
        else:
            index_2 = 3
            index_3 = 2

        box = [
            points[index_1], points[index_2], points[index_3], points[index_4]
        ]
        return box, min(bounding_box[1])

    def box_score_fast(self, bitmap, _box):
        h, w = bitmap.shape[:2]
        box = _box.copy()
        xmin = np.clip(np.floor(box[:, 0].min()).astype(np.int32), 0, w - 1)
        xmax = np.clip(np.ceil(box[:, 0].max()).astype(np.int32), 0, w - 1)
        ymin = np.clip(np.floor(box[:, 1].min()).astype(np.int32), 0, h - 1)
        ymax = np.clip(np.ceil(box[:, 1].max()).astype(np.int32), 0, h - 1)

        mask = np.zeros((ymax - ymin + 1, xmax - xmin + 1), dtype=np.uint8)
        box[:, 0] = box[:, 0] - xmin
        box[:, 1] = box[:, 1] - ymin
        cv2.fillPoly(mask, box.reshape(1, -1, 2).astype(np.int32), 1)
        return cv2.mean(bitmap[ymin:ymax + 1, xmin:xmax + 1], mask)[0]

    def __call__(self, outs_dict, ratio_list):
        pred = outs_dict['maps']

        pred = pred[:, 0, :, :]
        segmentation = pred > self.thresh

        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            height, width = pred.shape[-2:]
            tmp_boxes, tmp_scores = self.boxes_from_bitmap(pred[batch_index],
                                                           segmentation[batch_index],
                                                           width, height)
            boxes = []
            for k in range(len(tmp_boxes)):
                if tmp_scores[k] > self.box_thresh:
                    boxes.append(tmp_boxes[k])
            if len(boxes) > 0:
                boxes = np.array(boxes)

                ratio_h, ratio_w = ratio_list[batch_index]
                boxes[:, :, 0] = boxes[:, :, 0] / ratio_w
                boxes[:, :, 1] = boxes[:, :, 1] / ratio_h

            boxes_batch.append(boxes)
        return boxes_batch



import numpy as np
import sys
import paddle.fluid as fluid
import visionx.dbnet_crnn.tools.utility as utility
from visionx.dbnet_crnn.ppocr.db_process import DBProcessTest
from visionx.dbnet_crnn.ppocr.db_post_process import DBPostProcess


class TextDetector(object):
    def __init__(self, args, model_path):
        self.det_algorithm = args['det_algorithm']
        self.use_zero_copy_run = args['use_zero_copy_run']
        postprocess_params = {}
        if self.det_algorithm == "DB":
            self.preprocess_op = DBProcessTest()
            postprocess_params["thresh"] = args['det_db_thresh']
            postprocess_params["box_thresh"] = args['det_db_box_thresh']
            postprocess_params["max_candidates"] = 3000
            postprocess_params["unclip_ratio"] = args['det_db_unclip_ratio']
            self.postprocess_op = DBPostProcess(postprocess_params)
        else:
            print("unknown det_algorithm:{}".format(self.det_algorithm))
            sys.exit(0)

        self.predictor, self.input_tensor, self.output_tensors =\
            utility.create_predictor(args, mode="det", model_path=model_path)

    def order_points_clockwise(self, pts):
        """
        reference from: https://github.com/jrosebr1/imutils/blob/master/imutils/perspective.py
        # sort the points based on their x-coordinates
        """
        xSorted = pts[np.argsort(pts[:, 0]), :]

        # grab the left-most and right-most points from the sorted
        # x-roodinate points
        leftMost = xSorted[:2, :]
        rightMost = xSorted[2:, :]

        # now, sort the left-most coordinates according to their
        # y-coordinates so we can grab the top-left and bottom-left
        # points, respectively
        leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
        (tl, bl) = leftMost

        rightMost = rightMost[np.argsort(rightMost[:, 1]), :]
        (tr, br) = rightMost

        rect = np.array([tl, tr, br, bl], dtype="float32")
        return rect

    def clip_det_res(self, points, img_height, img_width):
        for pno in range(points.shape[0]):
            points[pno, 0] = int(min(max(points[pno, 0], 0), img_width - 1))
            points[pno, 1] = int(min(max(points[pno, 1], 0), img_height - 1))
        return points

    def filter_tag_det_res(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            box = self.order_points_clockwise(box)
            box = self.clip_det_res(box, img_height, img_width)
            rect_width = int(np.linalg.norm(box[0] - box[1]))
            rect_height = int(np.linalg.norm(box[0] - box[3]))
            if rect_width <= 10 or rect_height <= 10:
                continue
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def filter_tag_det_res_only_clip(self, dt_boxes, image_shape):
        img_height, img_width = image_shape[0:2]
        dt_boxes_new = []
        for box in dt_boxes:
            box = self.clip_det_res(box, img_height, img_width)
            dt_boxes_new.append(box)
        dt_boxes = np.array(dt_boxes_new)
        return dt_boxes

    def __call__(self, img, max_side_len):
        ori_im = img.copy()
        im, ratio_list = self.preprocess_op(img, max_side_len)
        if im is None:
            return None, 0
        im = im.copy()
        if self.use_zero_copy_run:
            self.input_tensor.copy_from_cpu(im)
            self.predictor.zero_copy_run()
        else:
            im = fluid.core.PaddleTensor(im)
            self.predictor.run([im])
        outputs = []
        for output_tensor in self.output_tensors:
            output = output_tensor.copy_to_cpu()
            outputs.append(output)
        outs_dict = {}
        if self.det_algorithm == "EAST":
            outs_dict['f_geo'] = outputs[0]
            outs_dict['f_score'] = outputs[1]
        elif self.det_algorithm == 'SAST':
            outs_dict['f_border'] = outputs[0]
            outs_dict['f_score'] = outputs[1]
            outs_dict['f_tco'] = outputs[2]
            outs_dict['f_tvo'] = outputs[3]
        else:
            outs_dict['maps'] = outputs[0]

        dt_boxes_list = self.postprocess_op(outs_dict, [ratio_list])
        dt_boxes = dt_boxes_list[0]
        dt_boxes = self.filter_tag_det_res(dt_boxes, ori_im.shape)
        return dt_boxes



#!/usr/bin/env python
# -*- coding:utf-8 -*-
from setuptools import setup, find_packages
from visionx import __version__

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


def parse_requirements(filename):
    """ load requirements from a pip requirements file. (replacing from pip.req import parse_requirements)"""
    lineiter = (line.strip() for line in open(filename))
    reqs = [line for line in lineiter if line and not line.startswith("#")]
    return reqs


setup(
    name="visionx",
    version="{}".format(__version__),
    keywords=["vision", "utx", "yolox", "diff", "ui", "tools"],
    description='Video UI analysis tool!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Apache License 2.0',

    url="https://github.com/openutx",
    author="lijiawei",
    author_email="jiawei.li2@qq.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console :: Curses",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython"
    ],
    entry_points="""
    [console_scripts]
    visionx = visionx.server:main
    """,
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=parse_requirements('requirements.txt')

)



# 配置文件
import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# intra_op_num_threads为CPU核数效能最佳
OP_NUM_THREADS = 4
IMAGE_INFER_MODEL_PATH = os.path.join(BASE_DIR, 'capture/local_models/ui_det_v2.onnx')
CLIP_MODEL_PATH = os.path.join(BASE_DIR, 'capture/local_models/clip_vit32_feat.onnx')
IMAGE_TEMP_DIR = os.path.join(BASE_DIR, 'capture/temp')
REC_CHAR_DICT_PATH = os.path.join(BASE_DIR, 'dbnet_crnn/ppocr/utils/keys.txt')
PREDICT_DET_PATH = os.path.join(BASE_DIR, 'dbnet_crnn/modelv1.1/det/')
PREDICT_REC_PATH = os.path.join(BASE_DIR, 'dbnet_crnn/modelv1.1/rec/')



import os.path
import re
import cv2
import numpy as np
import onnxruntime
import time
from visionx.config import IMAGE_INFER_MODEL_PATH, OP_NUM_THREADS
from visionx.service.image_utils import yolox_preprocess, yolox_postprocess, multiclass_nms, img_show


class ImageInfer(object):
    def __init__(self, model_path):
        self.UI_CLASSES = ("bg", "icon", "pic")
        self.input_shape = [640, 640]
        self.cls_thresh = 0.5
        self.nms_thresh = 0.2
        self.model_path = model_path
        so = onnxruntime.SessionOptions()
        so.intra_op_num_threads = OP_NUM_THREADS
        self.model_session = onnxruntime.InferenceSession(self.model_path, sess_options=so)

    def ui_infer(self, image):
        origin_img = cv2.imread(image) if isinstance(image, str) else image
        img, ratio = yolox_preprocess(origin_img, self.input_shape)
        ort_inputs = {self.model_session.get_inputs()[0].name: img[None, :, :, :]}
        output = self.model_session.run(None, ort_inputs)
        predictions = yolox_postprocess(output[0], self.input_shape)[0]
        boxes = predictions[:, :4]
        scores = predictions[:, 4:5] * predictions[:, 5:]
        boxes_xyxy = np.ones_like(boxes)
        boxes_xyxy[:, 0] = boxes[:, 0] - boxes[:, 2] / 2.
        boxes_xyxy[:, 1] = boxes[:, 1] - boxes[:, 3] / 2.
        boxes_xyxy[:, 2] = boxes[:, 0] + boxes[:, 2] / 2.
        boxes_xyxy[:, 3] = boxes[:, 1] + boxes[:, 3] / 2.
        boxes_xyxy /= ratio
        dets = multiclass_nms(boxes_xyxy, scores, nms_thr=self.nms_thresh, score_thr=self.cls_thresh)
        if dets is not None:
            # 兼容不同版本模型返回结果中UI classes index起始位置
            offset = 0
            match_obj = re.match(r'.*o(\d+)\.onnx$', self.model_path)
            if match_obj:
                offset = int(match_obj.group(1))
            dets[:, 5] += offset
        return dets

    def show_infer(self, dets, origin_img, infer_result_path):
        if dets is not None:
            boxes, scores, cls_inds = dets[:, :4], dets[:, 4], dets[:, 5]
            origin_img = img_show(origin_img, boxes, scores, cls_inds, conf=self.cls_thresh,
                                  class_names=self.UI_CLASSES)
        cv2.imwrite(infer_result_path, origin_img)


image_infer = ImageInfer(IMAGE_INFER_MODEL_PATH)


def get_ui_infer(image, cls_thresh):
    """
    elem_det_region x1,y1,x2,y2
    """
    data = []
    image_infer.cls_thresh = cls_thresh if isinstance(cls_thresh, float) else image_infer.cls_thresh
    dets = image_infer.ui_infer(image)
    if isinstance(dets, np.ndarray):
        boxes, scores, cls_inds = dets[:, :4], dets[:, 4], dets[:, 5]
        for i in range(len(boxes)):
            box = boxes[i]
            box[box < 0] = 0
            box = box.tolist() if isinstance(box, (np.ndarray,)) else box
            elem_type = image_infer.UI_CLASSES[int(cls_inds[i])]
            score = scores[i]
            data.append(
                {
                    "elem_det_type": "image" if elem_type == 'pic' else elem_type,
                    "elem_det_region": box,
                    "probability": score
                }
            )
    return data


if __name__ == '__main__':
    """
    调试代码
    """
    image_path = "./capture/image_1.png"
    infer_result_path = "./capture/local_images"
    assert os.path.exists(image_path)
    assert os.path.exists(IMAGE_INFER_MODEL_PATH)
    if not os.path.exists(infer_result_path):
        os.mkdir(infer_result_path)
    t1 = time.time()
    dets = image_infer.ui_infer(image_path)
    print(f"Infer time: {round(time.time()-t1, 3)}s")
    infer_result_name = f"infer_result.png"
    image_infer.show_infer(dets, cv2.imread(image_path), os.path.join(infer_result_path, infer_result_name))
    print(f"Result saved {infer_result_path}/{infer_result_name}")



from pprint import pprint

from visionx.service.image_utils import *
from visionx.dbnet_crnn.image_text import image_text


def get_image_text(image):
    img = cv2.imread('/Users/ymt/Documents/GitHub/utx-vision/visions/capture/' + image)
    h, w, _ = img.shape
    log_size = int(0.9*h)
    result = {
        'roi_text': image_text.get_text(img, log_size),
        'img_shape': img.shape
    }
    return result


if __name__ == '__main__':
    pprint(get_image_text('IMG_34237CB14429-1.jpeg'))